
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar
import matplotlib.transforms as mtransforms
import earthpy as et
import earthpy.spatial as es
import earthpy.plot as ep
import rasterio
import numpy as np
from rasterio.plot import plotting_extent
from matplotlib.colors import TwoSlopeNorm, Normalize
import matplotlib.patches as patches
import os
import rasterio
import glob
import seaborn as sb
import pandas as pd
import matplotlib.dates as mdates
import matplotlib as mpl
from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib
from matplotlib.collections import LineCollection
# Import needed packages
import geopandas as gpd
import rioxarray as rxr
import earthpy as et
import earthpy.spatial as es
import rasterio as rio
import cmocean
import plotly.graph_objects as go
import plotly.express as px
import plotly.express as px_temp
import seaborn as sns  
import plotly.offline as py_offline
import statsmodels.api as sm
from sklearn.metrics import mean_squared_error, r2_score
from sklearn.linear_model import LinearRegression
from datetime import datetime
import math
from ipywidgets import interact
from ipywidgets import widgets
import plotly.io as pio
import re





def akhdefo_viewer(Path_to_DEMFile="", rasterfile="", title='', pixel_resolution_meter=3.125,
                    outputfolder="", outputfileName="", alpha=0.5, unit='', noDATA_MAsk=False, cmap='jet', 
                    min_value=None, max_value=None, normalize=False, colorbar_label=None, show_figure=True):
    """
    This function overlays a raster file on a DEM hillshade and saves the plot as a .png image. 

    Parameters:
    Path_to_DEMFile (str): Path to the DEM file.
    rasterfile (str): Path to the raster file.
    title (str, optional): Title of the plot. If None, the basename of the raster file is used.
    pixel_resolution_meter (float, optional): Pixel resolution of the raster in meters. Default is 3.125.
    outputfolder (str): Path to the folder where the output image will be saved.
    outputfileName (str): Name of the output .png image.
    alpha (float, optional): Alpha value for the raster overlay. Default is 0.5.
    unit (int, optional): Unit conversion factor for the raster values. Default is ''. "100cm" if your data in meters want to convert to cm, adjust to your need e.g. "1000mm", etc..
    noDATA_MAsk (bool, optional): If True, pixels with a value of 0 in the raster are masked. Default is False.
    cmap (str, optional): Colormap to use for the raster. Default is 'jet'.
    min_value (float, optional): Minimum value to use for the normalization. If None, the minimum raster value is used.
    max_value (float, optional): Maximum value to use for the normalization. If None, the maximum raster value is used.
    normalize (bool, optional): If True, the raster values are normalized. Default is False.
    colorbar_label (str, optional): Label for the colorbar. If None, no label is added. 

    Returns:
    None
    """
    with rasterio.open(Path_to_DEMFile) as src:
        dem = src.read(1)
        dem_transform = src.transform

    hillshade = es.hillshade(dem)

    with rasterio.open(rasterfile) as src:
        raster = src.read(1)
        raster_transform = src.transform
        raster_crs = src.crs

    if noDATA_MAsk:
        raster = np.ma.masked_where(raster == 0, raster)

    
    def separate_floats_letters(input_string):
        floats = re.findall(r'\d+\.\d+|\d+', input_string)
        letters = re.findall(r'[a-zA-Z]+', input_string)
        return letters, floats

    input_string = unit
    letters, unit_val = separate_floats_letters(input_string)
    
    if  unit:
        raster = raster * float(unit_val)

    if normalize:
        if min_value is None:
            min_value = raster.min()
        if max_value is None:
            max_value = raster.max()

        if raster.min() < 0:
            norm = TwoSlopeNorm(vmin=min_value, vcenter=0, vmax=max_value)
        else:
            norm = Normalize(vmin=min_value, vmax=max_value)
    else:
        norm = None

    fig, ax = plt.subplots(figsize=(7, 7))

    # Plot the hillshade layer
    ep.plot_bands(
        hillshade,
        ax=ax,
        cmap='gray',
        scale=False,
        cbar=False,
        extent=plotting_extent(src)
    )

    if not title:
        title = os.path.splitext(os.path.basename(rasterfile))[0]
    if not outputfileName:
        outputfileName=os.path.splitext(os.path.basename(rasterfile))[0]

    # Overlay the raster with alpha for transparency
    img = ax.imshow(raster, alpha=alpha, cmap=cmap, norm=norm,
                    extent=plotting_extent(src))

    if colorbar_label:
        cbar_ax = fig.add_axes([0.92, 0.22, 0.02, 0.5])
        fig.colorbar(img, cax=cbar_ax, label=colorbar_label, extend='both')

    if raster_crs.is_geographic:
        ax.set_xlabel('Longitude')
        ax.set_ylabel('Latitude')
    else:
        ax.set_xlabel('Easting')
        ax.set_ylabel('Northing')

    ax.grid(True, which='major')
    ax.set_title(title)

    scalebar = AnchoredSizeBar(ax.transData, 5000, f'{5000 * pixel_resolution_meter} m', 'lower right',
                               pad=0.6,
                               color='black',
                               frameon=True,
                               size_vertical=1)
    ax.add_artist(scalebar)

    plt.savefig(os.path.join(outputfolder, outputfileName), dpi=100, bbox_inches='tight')
    
    if show_figure==True:
        plt.show()
    else:
        plt.close()


def plot_stackNetwork(src_folder=r"", output_folder=r"" , cmap='tab20', date_plot_interval=(5, 30), marker_size=15):
    '''
    This Program plots temporal network of triplets to be stacked for calculating 
    Annual Mean Velocity from stacked optical images.

    Parameters
    ----------
    src_folder: str
        path to georeferenced_folder

    output_folder: str
        path to output folder to save output Figure plot

    cmap: str
        colormap for the plot default is "tab20"

    date_plot_interval: list
        minumum and maximum plot x axis interval dates for the plot
    
    marker_size: float
        size of plotted points default is 15

    Returns
    -------
    Figure
    '''
    
    
    if not os.path.exists(output_folder ):
        os.makedirs(output_folder)

    #20200218-20200220-20200226.tif

    path_folder=sorted(glob.glob( src_folder + "/" + "flow_xn/*.tif"))
    df=pd.DataFrame(columns= ["Triplet_Dates", "Triplet ID", "Time"])
    label_date_list=[]
    label_time_list=[]
    label_id_list=[]
    delta_dd_list=[]
    for idx, item in enumerate(path_folder):
        
        filepath,filename=os.path.split(path_folder[idx])
        label_date=filename[:-4]
        label_time=filename[:-22]
        label_date_list.append(label_date)
        label_time_list.append(label_time)
        label_id_list.append(idx)

        ###
        #20200218-20200220-20200226.tif
        Date1_YYYY=filename[:-26]
        Date1_MM=filename[4:-24]
        Date1_DD=filename[6:-22]

        Date2_YYYY=filename[9:-17]
        Date2_MM=filename[13:-15]
        Date2_DD=filename[15:-13]

        Date3_YYYY=filename[18:-8]
        Date3_MM=filename[22:-6]
        Date3_DD=filename[24:-4]
        #convert dates to number of days in the year for image1
        YMD= Date1_YYYY+Date1_MM+Date1_DD
        date1 = pd.to_datetime(YMD, format='%Y%m%d')
        new_year_day = pd.Timestamp(year=date1.year, month=1, day=1)
        day_of_the_year_date1 = (date1 - new_year_day).days + 1

        #convert dates to number of days in the year for image2
        YMD= Date2_YYYY+Date2_MM+Date2_DD
        date2 = pd.to_datetime(YMD, format='%Y%m%d')
        new_year_day = pd.Timestamp(year=date2.year, month=1, day=1)
        day_of_the_year_date2 = (date2 - new_year_day).days + 1

            #convert dates to number of days in the year for image3
        YMD= Date3_YYYY+Date3_MM+Date3_DD
        date3 = pd.to_datetime(YMD, format='%Y%m%d')
        new_year_day = pd.Timestamp(year=date3.year, month=1, day=1)
        day_of_the_year_date3 = (date3 - new_year_day).days + 1

        Delta_DD= (date3-date1).days
        Delta_DD=int(Delta_DD)
        if Delta_DD < 0:
            Delta_DD=Delta_DD*-1
        else:
            Delta_DD=Delta_DD*1
        delta_dd_list.append(Delta_DD)   
        #print (delta_dd_list)
    df["Triplet_Dates"]=(label_date_list)
    df["Triplet ID"]=(label_id_list)
    df["Time"]=(label_time_list)
    df['Delta_DD']=delta_dd_list
    df['Time'] = pd.to_datetime(df['Time'].astype(str), format='%Y%m%d')
   
    t=np.array(df["Time"])
    lb=np.array(df["Triplet_Dates"])
    lb_id=np.array(df["Triplet ID"])
    delta_dates=np.array(df['Delta_DD'])
    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(10,5))
    fig.autofmt_xdate()
    a=ax.scatter( t, delta_dates, c=mdates.date2num(t), cmap=cmap, s=marker_size , norm=matplotlib.colors.Normalize())
    ax.plot(t, delta_dates , color='k' , alpha=0.5)
    
    cb=fig.colorbar(a, ax=ax, orientation='horizontal', pad=0.02)
    loc_major = mdates.AutoDateLocator(minticks=date_plot_interval[0], maxticks=date_plot_interval[1])
    #loc_minor = mdates.AutoDateLocator(minticks=1, maxticks =5)
    #cb.ax.xaxis.set_minor_locator(loc_minor)
    cb.ax.xaxis.set_major_locator(loc_major)
    #cb.ax.xaxis.set_minor_formatter(mdates.ConciseDateFormatter(loc_minor))
    cb.ax.xaxis.set_major_formatter(mdates.ConciseDateFormatter(loc_major))
    plt.gca().xaxis.set_major_locator(mdates.DayLocator(interval=date_plot_interval[1])) 
    #plt.gca().xaxis.set_minor_locator(mdates.DayLocator(interval=date_plot_interval[0])) 
    plt.xlim(df['Time'][df.index[0]],df['Time'][df.index[-1]] )
    #plt.title("Triplet Stack Network")
    #plt.xlabel("Dates")
    plt.ylabel("Number of Days for Each Triplet")
    # handles, labels = a.legend_elements(prop='colors', alpha=0.2)
    # legend = ax.legend(handles, labels, bbox_to_anchor=(1.08, 1), loc="upper center", title="Number of days" )
    
    cb.ax.tick_params(rotation=90)
    ax.xaxis.tick_top()
      
    ax.xaxis.set_label_position('top') 
    ax.set_xlabel('Dates') 
    plt.xticks(rotation=85) 
    ax.grid(True)

    #print(df.head(5))
    
    plt.savefig( output_folder + "/" + "Stack_Network.jpg", dpi=300, bbox_inches = 'tight')



    

def akhdefo_ts_plot(path_to_shapefile=r"", dem_path=r"", point_size=1.0, opacity=0.75, cmap="turbo",
                    Set_fig_MinMax=True, MinMaxRange=[-50,50] , color_field='VEL', user_data_points="", 
                    path_saveData_points="" , save_plot=False, Fig_outputDir='' , VEL_Scale='year', filename_dates=''):
    '''
    This program used for analysis time-series velocity profiles

    Parameters
    ----------

    user_data_points: str
        provide path to csv. file contains x and y coordinate for points of interest
        you can generate this file by providing path to path_saveData_points (POI.csv).
        This is useful to save mouse click positions to repeat the plots for different datasets for example if you plot several TS profiles for
        EW velocity product, you can recreate TS for the same exact position by saving POI.csv with path_saveData_points and then use that as input for the another
        plot such as NS velocity product via setting user_datapoints="POI.csv"
    
    path_to_shapefile: str 
        type path to timeseries shapefile in stack_data/TS folder
    
    dem_path: str
        path to dem raster in geotif fromat

    point_size: float
        size of the sactter plot points

    opacity: float 
        transparency of the scater overlay

    cmap: str
        Matplotlib colormap options example "RdYlBu_r, jet, turbo, hsv, etc..."
                    
    Set_fig_MinMax: bool
        True or False

    MinMaxRange: list
        [-50,50]  Normalize plot colormap range if Set_fig_MinMax=True

    color_field: str 
        'VEL' ,"VEL_2D", 'VEL_N', 'VEL_E', 'VELDir_MEA'

   
    path_saveData_points: str
        optional, provide directory path if you want to save profile data.
        the data will be saved under POI.csv file


    save_plot: bool
        True or False

    Fig_outputDir: str
        if save_plot=True then
        you save your profile plots in interactive html file and jpg image 

    VEL_Scale: str
        'year' or 'month' projects the velocity into provided time-scale

    filename_dates: str
        provide path to Names.txt file, this file generated at stack_prep step
        
        
    Returns
    -------
    Interactive Figures
    
    '''
     #####################################################################
    
    #################################################
    def ts_plot(df, plot_number, save_plot=False , output_dir="", plot_filename="" , VEL_Scale=VEL_Scale):
    
       
        pio.renderers.default = 'plotly_mimetype'
        
        df=pd.read_csv("temp.csv")
        

        df.rename(columns={ df.columns[0]: "dd" }, inplace = True)
        df['dd_str']=df['dd'].astype(str)
        df['dd_str'] = df['dd_str'].astype(str)
        df.rename(columns={ df.columns[1]: "val" }, inplace = True)
        df['dd']= pd.to_datetime(df['dd'].astype(str), format='%Y%m%d')
        
        df=df.set_index('dd')
        
        ########################
        df=df.dropna()
        # Make index pd.DatetimeIndex
        df.index = pd.DatetimeIndex(df.index)
        # Make new index
        idx = pd.date_range(df.index.min(), df.index.max())
        # Replace original index with idx
        df = df.reindex(index = idx)
        # Insert row count
        df.insert(df.shape[1],
                'row_count',
                df.index.value_counts().sort_index().cumsum())

        df=df.dropna()
        
        #df=df.set_index(df['row_count'], inplace=True)

        df.sort_index(ascending=True, inplace=True)
    
    
    
    
        
        
        #####start building slider
        widgets.SelectionRangeSlider(
        options=df.index,
        description='Dates',
        orientation='horizontal',
        layout={'width': '1000px'})
        
        
        ############
        def ts_helper(df, VEL_Scale=VEL_Scale, plot_number=plot_number):
            

            def best_fit_slope_and_intercept(xs,ys):
                from statistics import mean
                xs = np.array(xs, dtype=np.float64)
                ys = np.array(ys, dtype=np.float64)
                m = (((mean(xs)*mean(ys)) - mean(xs*ys)) /
                    ((mean(xs)*mean(xs)) - mean(xs*xs)))
                
                b = mean(ys) - m*mean(xs)
                
                return m, b

            #convert dattime to number of days per year
            
            
            dates_list=([datetime.strptime(x, '%Y%m%d') for x in df.dd_str])
            #days_num=[( ((x) - (pd.Timestamp(year=x.year, month=1, day=1))).days + 1) for x in dates_list]
            dd_days=dates_list[len(dates_list)-1]- dates_list[0]
            print(dates_list[0], dates_list[len(dates_list)-1] , dd_days)
            dd_days=str(dd_days)
            dd_days=dd_days.removesuffix('days, 0:00:00')
            delta=int(dd_days)
            m, b = best_fit_slope_and_intercept(df.row_count, df.val)
            print("m:", math.ceil(m*100)/100, "b:",math.ceil(b*100)/100)
            regression_model = LinearRegression()
            val_dates_res = regression_model.fit(np.array(df.row_count).reshape(-1,1), np.array(df.val))
            y_predicted = regression_model.predict(np.array(df.row_count).reshape(-1,1))
        
            if VEL_Scale=='year':
                rate_change=regression_model.coef_[0]/delta * 365.0
                std=np.std(y_predicted)/delta * 365
            elif VEL_Scale=='month':
                rate_change=regression_model.coef_[0]/delta * 30.0
                std=np.std(y_predicted)/delta * 30
            else:
                rate_change=regression_model.coef_[0]  #delta is number of days
                std=np.std(y_predicted)
                
            # model evaluation
            mse=mean_squared_error(np.array(df.val),y_predicted)
            rmse = np.sqrt(mean_squared_error(np.array(df.val), y_predicted))
            r2 = r2_score(np.array(df.val), y_predicted)
            
            # printing values
            slope= ('Slope(linear deformation rate):' + str(math.ceil((regression_model.coef_[0]/delta)*100)/100) + " mm/day")
            Intercept=('Intercept:'+ str(math.ceil(b*100)/100))
            #print('MSE:',mse)
            rmse=('Root mean squared error: '+ str(math.ceil(rmse*100)/100))
            r2=('R2 score: '+ str(r2))
            
            std=("STD: "+ str(math.ceil(std*100)/100)) 
            # Create figure
            #fig = go.Figure()
            
            return y_predicted, rate_change, slope, Intercept, rmse, r2, std, plot_number, print(len(df)), dd_days
        
        
        @interact
        def read_values(
            slider = widgets.SelectionRangeSlider(
            options=df.index,
            index=(0, len(df.index) - 1),
            description='Dates',
            orientation='horizontal',
            layout={'width': '500px'},
            continuous_update=True) ):
            
            #df=pd.read_csv("temp.csv")
            df=pd.read_csv("temp.csv")

            df.rename(columns={ df.columns[0]: "dd" }, inplace = True)
            df['dd_str']=df['dd'].astype(str)
            df['dd_str'] = df['dd_str'].astype(str)
            df.rename(columns={ df.columns[1]: "val" }, inplace = True)
            df['dd']= pd.to_datetime(df['dd'].astype(str), format='%Y%m%d')
            
            df=df.set_index('dd')
            
            ########################
            df=df.dropna()
            # Make index pd.DatetimeIndex
            df.index = pd.DatetimeIndex(df.index)
            # Make new index
            idx = pd.date_range(df.index.min(), df.index.max())
            # Replace original index with idx
            df = df.reindex(index = idx)
            # Insert row count
            df.insert(df.shape[1],
                    'row_count',
                    df.index.value_counts().sort_index().cumsum())

            df=df.dropna()
            
            #df=df.set_index(df['row_count'], inplace=True)

            df.sort_index(ascending=True, inplace=True)
            
            
            
            df=df.loc[slider[0]: slider[1]]
            
            
             
            
            helper=ts_helper(df, VEL_Scale=VEL_Scale)
            
            y_predicted=helper[0]
            rate_change=helper[1]
            slope=helper[2]
            Intercept=helper[3]
            rmse=helper[4]
            r2=helper[5]
            std=helper[6]
            plot_number=helper[7]
            
            print(rate_change)
            print(slope)
            print(rmse)
            print(std)
            print(Intercept)
            
        

            fig = go.Figure()
            fig.update_xaxes(range=[slider[0], slider[1]])
            trace1 = go.Scatter(x=(df.index), y=(y_predicted), mode='lines', name='Trendline')
            fig.add_trace(trace1)
            trace2 = go.Scatter(x=(df.index), y=(df.val), mode='markers', name='Data-Points')
            fig.add_trace(trace2)
            trace3 = go.Scatter(x=(df.index), y=(df.val), mode='lines', name='Draw-line', visible='legendonly')
            fig.add_trace(trace3)
            
            fig.update_layout(xaxis_title="Date", yaxis_title="millimeter")
            
            unit=helper[9]+ "days"
            if VEL_Scale=="year":
                unit="year"
            elif VEL_Scale=="month":
                unit="month"
            else:
                unit=unit
            
            tt= "Defo-Rate:"+str(round(rate_change,2))+"mm/"+unit+":"+ "Defo-Rate-STD:"+str(round(np.std(y_predicted), 2))+ ":Plot ID-" + str(plot_number)
            
            fig.update_layout(
                
            title_text=tt, title_font_family="Sitka Small",
            title_font_color="red", title_x=0.5 , legend_title="Legend",
            font=dict(
                family="Courier New, monospace",
                size=15,
                color="RebeccaPurple" ))
            
            fig.update_layout(font_family="Sitka Small")
            
            # fig.update_layout(legend=dict(
            # yanchor="top",
            # y=-0,
            # xanchor="left",
            # x=1.01))
            fig.update_xaxes(tickformat='%Y.%m.%d')
            fig.update_layout(xaxis = go.layout.XAxis( tickangle = 45))
            
            
            fig.update_layout(hovermode="x unified")
            
           
            f2=go.FigureWidget(fig.to_dict()).show()
           
            
            if save_plot==True:
            
                if not os.path.exists(output_dir):
                    os.mkdir(output_dir)

                fig.write_html(output_dir + "/" + plot_filename + ".html" )
                fig.write_image(output_dir + "/" + plot_filename + ".jpeg", scale=1, width=1080, height=300 )
            
            f2
     
    
   #######################################################################3
   # Define the file name to check
    filename_temp_dates = filename_dates 
    # Check if the file exists in the current working directory
    if os.path.exists(filename_temp_dates):
        filename_dates=filename_temp_dates
        #print(f'The file "{filename_dates}" exists in the current directory.')
    else:
        print(f'The file "{filename_dates}" does not exist in the current directory.\n Please enter path to Names.txt to filename_dates variable')
    
    dnames=[]
    with open(filename_dates, 'r') as fp:
        for line in fp:
            # remove linebreak from a current name
            # linebreak is the last character of each line
            x =  'D' + line[:-1]

            # add current item to the list
            dnames.append(x[:-18])

   
    
    # Import shapfilepath
    shapefile_path = os.path.join(path_to_shapefile)
    basename = os.path.basename(shapefile_path[:-4])
    # Open shapefile data with geopandas
    gdf = gpd.read_file(shapefile_path)
    gdf.crs
    # Define path to dem data
    #dem_path = 'dem.tif'

    with rio.open(dem_path) as src:
        elevation = src.read(1)
        elevation = elevation.astype('float32')
        # Set masked values to np.nan
        elevation[elevation < 0.0] = np.nan
        
    # Create and plot the hillshade with earthpy
    hillshade = es.hillshade(elevation, azimuth=270, altitude=45)

    dem = rxr.open_rasterio(dem_path, masked=True)
    dem_plotting_extent = plotting_extent(dem[0], dem.rio.transform())

    # Getting the crs of the raster data
    dem_crs = es.crs_check(dem_path)
    

    # Transforming the shapefile to the dem data crs
    gdf = gdf.to_crs(dem_crs)
    
    
    min=gdf[color_field].min()
    max=gdf[color_field].max()
    import matplotlib.colors as mcolors
    from matplotlib_scalebar.scalebar import ScaleBar
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    
    
    if min < 0 and max >0 :
        if Set_fig_MinMax==True:
            min_n=MinMaxRange[0]
            max_n=MinMaxRange[1]
            min=min_n
            max=max_n
            offset = mcolors.TwoSlopeNorm(vmin=min,
                        vcenter=0., vmax=max)  
        else:
            offset = mcolors.TwoSlopeNorm(vmin=min,
                        vcenter=0., vmax=max)
            
    else  : 
        if Set_fig_MinMax==True:
            min_n=0
            max_n=100
            min=MinMaxRange[0]
            max=MinMaxRange[1]
            offset=mcolors.Normalize(vmin=min, vmax=max)
        else:
            offset=mcolors.Normalize(vmin=min, vmax=max)


    if user_data_points!="":
        print("Click Anywhere on the Figure to plot Your Time-Series Profile and plot Point Locations")


    fig = plt.figure(figsize=(7,7))
    ax1 = fig.add_subplot(111)
    
    divider = make_axes_locatable(ax1)
    cax = divider.append_axes('bottom', size='5%', pad=0.5)
    
    #fig, ax1 = plt.subplots(ncols=1, nrows=1, figsize=(10,5))
    ep.plot_bands( hillshade,cbar=False,title=basename,extent=dem_plotting_extent,ax=ax1, scale=False)
    img_main=ax1.scatter(gdf.x, gdf.y, c=gdf[color_field], alpha=opacity, s=point_size, picker=1, cmap=cmap, norm=offset)
    scalebar = ScaleBar(1, "m", length_fraction=0.25, scale_loc="right",border_pad=1,pad=0.5, box_color='white', box_alpha=0.5, location='lower right')
    ax1.add_artist(scalebar)
    plt.grid(True)
    #ax.scatter(gdf.x, gdf.y, s= 0.5, c=gdf.VEL_MEAN ,picker=1)
    cb=fig.colorbar(img_main, ax=ax1, cax=cax, extend='both', orientation='horizontal')
    cb.set_label('mm/year', labelpad=2, x=0.5, rotation=0)
    
    global count
    count=0
    
    x_list=[]
    y_list=[]
    label_ID_list=[]
    df_filt1_list=[]
    
    
    def onclick(event):
        global count
        count+=1
        
        global ix, iy
        ix, iy = event.xdata, event.ydata
        print('button=%d, Figure Coordinates: x=%d, y=%d, : Geographic Coordinates: xdata=%f, ydata=%f' % 
            (event.button, event.x, event.y, event.xdata, event.ydata))
        #plt.plot(event.xdata, event.ydata, ',')
        def find_nearest(array, value):
            array = np.asarray(array)
            idx = (np.abs(array - value)).argmin()
            return array[idx]
         
        def filter_rowwise(gdf,find_nearest, ix, iy ):  
            s1=find_nearest(np.array(gdf.x), ix)
            s2=find_nearest(np.array(gdf.y), iy)
            cols = ['x', 'y']
            vals = [s1, s2]
            geo_df=gdf[np.logical_and.reduce([gdf[c] == v for c, v in zip(cols, vals)])]
            df_filt1=geo_df[dnames]
            
            df_filt1.columns = df_filt1.columns.str.replace(r"D", "")
            df_filt1_list.append(df_filt1)
            
            df=df_filt1.T
            
            df.to_csv('temp.csv')
            
            return s1, s2
        
        
        # ##################
        # def pass_flag(gdf):
        #     Flag="yes"
        #     for idx, row in gdf.iterrows():
        #         #print("Index:", idx, ": ",  row['x'], row['y']) 
        #         s1, s2=filter_rowwise(gdf,find_nearest, row['x'], row['y'] ) 
        #         if s1!=row['x'] and s2!=row['y']:
        #             print("No Data Available at this Location, Please click on the data points")
        #             Flag="no"
        #     return Flag
        # ########################3
        # flag=pass_flag(gdf)
        # if flag=='no':
        #     print("No Data Available at this Location, Please click on the data points")
          
        if user_data_points!="":
            df_poi=pd.read_csv(user_data_points)
            for idx, row in df_poi.iterrows():
                #print("Index:", idx, ": ",  row['x'], row['y'])
                
                s1, s2=filter_rowwise(gdf,find_nearest, row['x'], row['y'] )
                
                
                
                    
                df=pd.read_csv('temp.csv')
                ps=ts_plot(df, idx+1, save_plot=save_plot, output_dir=Fig_outputDir, plot_filename=basename+"_"+str(idx+1), VEL_Scale=VEL_Scale)
                ax1.scatter(s1, s2,  marker=idx+1,  label=idx+1, s=100)
                ax1.text( s1, s2, idx+1, fontdict=dict(color="black"),
                bbox=dict(facecolor="white",alpha=0.75))
                
        else:
            s1, s2=filter_rowwise(gdf,find_nearest, ix, iy )
            
        
            df=pd.read_csv('temp.csv')
            
            ps=ts_plot(df, count, save_plot=save_plot, output_dir=Fig_outputDir, plot_filename=basename+"_"+str(count), VEL_Scale=VEL_Scale)
            
            print("count: " , count)

    
        ########################33
            
            #os.unlink("temp.csv")
            
            
            x_list.append(s1)
            y_list.append(s2)
            label_ID_list.append(count)
            
            
            ax1.scatter(event.xdata, event.ydata,  marker=count,  label=count, s=100)
            ax1.text( event.xdata, event.ydata, count, fontdict=dict(color="black"),
                bbox=dict(facecolor="white",alpha=0.75))
        
        
        
        if path_saveData_points!="": 
            if not os.path.exists(path_saveData_points):
                os.mkdir(path_saveData_points)
            df_filt1=pd.concat(df_filt1_list)
            df_filt1['x']=x_list
            df_filt1['y']=y_list
            df_filt1['ID']=label_ID_list
            
            
            df_filt1 = df_filt1.loc[:, ~df_filt1.columns.str.contains('Unnamed')]
            cols = list(df_filt1.columns)
            cols = [cols[-1]] + cols[:-1]
            cols = [cols[-1]] + cols[:-1]
            cols = [cols[-1]] + cols[:-1]
            df_filt1 = df_filt1[cols]
            
            df_filt1.to_csv(path_saveData_points + "/" + "POI.csv")
            
        ax1.legend(loc='upper left') 
        if save_plot==True:
    
            if not os.path.exists(Fig_outputDir):
                os.mkdir(Fig_outputDir)
            
            plt.savefig(Fig_outputDir + "/" + basename + ".png" )
                
        #os.unlink("temp.csv")
          
    
    cid = fig.canvas.mpl_connect('button_press_event', onclick)
    
    
    
    
   

    
    
    
    plt.show() 
   

def MeanProducts_plot_ts(path_to_shapefile="", dem_path="" , out_folder="Figs_analysis", color_field="", Set_fig_MinMax=False, MinMaxRange=[-100,100],
                   opacity=0.5, cmap="jet" , point_size=1, cbar_label="mm/year" , batch_plot=False, dates_list="" ):
    
    """
    This program used to plot shapefile data

    Parameters
    ----------

    path_to_shapefile : str

    dem_path: str 

    out_folder: str

    color_field: str
        geopandas column name

    Set_fig_MinMax: bool

    MinMaxRange: list
            
    opacity: float

    cmap: str

    point_size: str 

    cbar_label: str
        "mm/year" or "degrees", etc.. based on unit of the data column name in the color_field

    Returns
    -------
    Figure
    """
   
    
    # Import shapfilepath
    shapefile_path = os.path.join(path_to_shapefile)
    basename = os.path.basename(shapefile_path[:-4])
    # Open shapefile data with geopandas
    gdf = gpd.read_file(shapefile_path)
    gdf.crs
    # Define path to dem data
    #dem_path = 'dem.tif'

    with rio.open(dem_path) as src:
        elevation = src.read(1)
        elevation = elevation.astype('float32')
        # Set masked values to np.nan
        elevation[elevation < 0.0] = np.nan
    # Create and plot the hillshade with earthpy
    hillshade = es.hillshade(elevation, azimuth=275, altitude=30)

    dem = rxr.open_rasterio(dem_path, masked=True)
    dem_plotting_extent = plotting_extent(dem[0], dem.rio.transform())

    # Getting the crs of the raster data
    dem_crs = es.crs_check(dem_path)
    

    # Transforming the shapefile to the dem data crs
    gdf = gdf.to_crs(dem_crs)
    
    if not os.path.exists(out_folder):
        os.makedirs(out_folder)
    
    if batch_plot==False:
        min=gdf[color_field].min()
        max=gdf[color_field].max()
        import matplotlib.colors as mcolors
        from matplotlib_scalebar.scalebar import ScaleBar
        from mpl_toolkits.axes_grid1 import make_axes_locatable
        
        
        if min < 0 and max >0 :
            if Set_fig_MinMax==True:
                min_n=MinMaxRange[0]
                max_n=MinMaxRange[1]
                min=min_n
                max=max_n
                offset = mcolors.TwoSlopeNorm(vmin=min,
                            vcenter=0., vmax=max)  
            else:
                offset = mcolors.TwoSlopeNorm(vmin=min,
                            vcenter=0., vmax=max)
                
        else  : 
            if Set_fig_MinMax==True:
                min_n=0
                max_n=100
                min=MinMaxRange[0]
                max=MinMaxRange[1]
                offset=mcolors.Normalize(vmin=min, vmax=max)
            else:
                offset=mcolors.Normalize(vmin=min, vmax=max)
                
        fig = plt.figure(figsize=(7,7))
        ax1 = fig.add_subplot(111)
        
        divider = make_axes_locatable(ax1)
        cax = divider.append_axes('bottom', size='5%', pad=0.5)
        
        #fig, ax1 = plt.subplots(ncols=1, nrows=1, figsize=(10,5))
        ep.plot_bands( hillshade,cbar=False,title=color_field,extent=dem_plotting_extent,ax=ax1, scale=False)
        img_main=ax1.scatter(gdf.x, gdf.y, c=gdf[color_field], alpha=opacity, s=point_size, picker=1, cmap=cmap, norm=offset)
        scalebar = ScaleBar(1, "m", length_fraction=0.25, scale_loc="right",border_pad=1,pad=0.5, box_color='white', box_alpha=0.5, location='lower right')
        ax1.add_artist(scalebar)
        ax1.grid(True)
        #ax.scatter(gdf.x, gdf.y, s= 0.5, c=gdf.VEL_MEAN ,picker=1)
        cb=fig.colorbar(img_main, ax=ax1, cax=cax, extend='both', orientation='horizontal')
        cb.set_label(cbar_label, labelpad=2, x=0.5, rotation=0)
        
        
        plt.savefig(out_folder+"/"+color_field+".png")
        
        plt.show()
    
    if batch_plot==True:
        if dates_list=="":
            print("provide list of dates in txt file(Names.txt file)")
        else:
            dnames=[]
            with open(dates_list, 'r') as fp:
                for line in fp:
                    # remove linebreak from a current name
                    # linebreak is the last character of each line
                    x = "D" + line[:-1]
                    # add current item to the list
                    dnames.append(x[:-18])
            
            for nd in gdf[dnames]:
                min=gdf[nd].min()
                max=gdf[nd].max()
                import matplotlib.colors as mcolors
                from matplotlib_scalebar.scalebar import ScaleBar
                from mpl_toolkits.axes_grid1 import make_axes_locatable
                
                
                if min < 0 and max >0 :
                    if Set_fig_MinMax==True:
                        min_n=MinMaxRange[0]
                        max_n=MinMaxRange[1]
                        min=min_n
                        max=max_n
                        offset = mcolors.TwoSlopeNorm(vmin=min,
                                    vcenter=0., vmax=max)  
                    else:
                        offset = mcolors.TwoSlopeNorm(vmin=min,
                                    vcenter=0., vmax=max)
                        
                else  : 
                    if Set_fig_MinMax==True:
                        min_n=0
                        max_n=100
                        min=MinMaxRange[0]
                        max=MinMaxRange[1]
                        offset=mcolors.Normalize(vmin=min, vmax=max)
                    else:
                        offset=mcolors.Normalize(vmin=min, vmax=max)
                fig = plt.figure(figsize=(7,7))
                ax1 = fig.add_subplot(111)
                
                divider = make_axes_locatable(ax1)
                cax = divider.append_axes('bottom', size='5%', pad=0.05)
                
                #fig, ax1 = plt.subplots(ncols=1, nrows=1, figsize=(10,5))
                ep.plot_bands( hillshade,cbar=False,title=nd,extent=dem_plotting_extent,ax=ax1, scale=False)
                img_main=ax1.scatter(gdf.x, gdf.y, c=gdf[nd], alpha=opacity, s=point_size, picker=1, cmap=cmap, norm=offset)
                scalebar = ScaleBar(1, "m", length_fraction=0.25, scale_loc="right",border_pad=1,pad=0.5, box_color='white', box_alpha=0.5, location='lower right')
                ax1.add_artist(scalebar)
                plt.grid(True)
                #ax.scatter(gdf.x, gdf.y, s= 0.5, c=gdf.VEL_MEAN ,picker=1)
                cb=fig.colorbar(img_main, ax=ax1, cax=cax, extend='both', orientation='horizontal')
                cb.set_label(cbar_label, labelpad=2, x=0.5, rotation=0)
                
                
                plt.savefig(out_folder+"/"+nd+".png")
                
                if len(dnames)>10:
                    print("akhdefo is plotting more than 10 figures to avoid crushing python kernel we skip displaying figures. \n Please see figures inside provided out_folder path")
                else:
                    
                    plt.show()
                
            
            



def Time_Series(stacked_raster_EW=r"", stacked_raster_NS=r"", velocity_points=r"", dates_name=r"", output_folder="", outputFilename="",
                 std=1, VEL_Scale='year' , velocity_mode="mean", master_reference=False):
    
    '''
    This program uses candiate velocity points from stackprep function and performs linear interpolation in time-domain to calibrate
    stacked velocity. Additionally produces corrected timeseries velocity(daily) in a shapefile.
    
    Parameters
    ----------
    b   
    stacked_raster_EW: str
    
    stacked_raster_NS: str
    
    velocity_points: str 
        Velcity Candidate points
    
    dates_name: str
        text file include name of each date in format YYYYMMDD
    
    output_folder: str
    
    outputFilename: str
    
    VEL_Scale: str
        'year' , "month" or empty  to calculate velocity within provided dataset date range
    
    velocity_mode: str
        "mean" or "linear"
        
    master_reference: bool
        True if calculate TS to a single reference date, False if calculate TS to subsequent Reference dates
    
    Returns
    -------
    
    Time-series shape file of velocity and direction EW, NS, and 2D(resultant Velocity and direction)
    
    '''
    import glob
    import os
    from datetime import datetime
    from os.path import join

    import dask.dataframe as dd
    import geopandas as gpd
    import geowombat as gw
    import numpy as np
    import pandas as pd
    import scipy.stats as stats
    from dateutil import parser 
    
    
    def Helper_Time_Series(stacked_raster=r"", velocity_points=r"", dates_name=r"", output_folder="", outputFilename="", std=1 , VEL_Scale=VEL_Scale):
        
        '''
        stacked_raster: Path to raster stack .tif file
        
        velocity_points: Path to velocity points in arcmap shapfile format .shp
        
        dates_name: path to text file contains date names of each time series triplets .txt
        
        output_folder: Path to output Folder
        
        outputFilename: name for out time-series velocity shapefile
        '''
        
        
        
        if not os.path.exists(output_folder):
                os.makedirs(output_folder)
    
        
        #Open Raster stack, extract pixel info into shape file

        with gw.open(stacked_raster, stack_dim='time') as src:
            print(src)
            #df = src.gw.extract(velocity_points)
            df = src.gw.extract(velocity_points, use_client=True , dtype='float32')
            df[df.select_dtypes(np.float64).columns] = df.select_dtypes(np.float64).astype(np.float32)

        

        #Import names to label timeseries data    
        names = []
        dnames=[]
        with open(dates_name, 'r') as fp:
            for line in fp:
                # remove linebreak from a current name
                # linebreak is the last character of each line
                x = 'D'+ line[:-1]

                # add current item to the list
                names.append(x)
                dnames.append(x[:-18])

        print (len(dnames))
        print(len(df.columns))

        cci=(len(df.columns)- len(dnames))
        df2=df.iloc[:, cci:]
        cc=np.arange(1,cci)
        #Add Timesereises column names
        
        # #find outliers using z-score iter 1
        # lim = np.abs((df2[cc] - df2[cc].mean(axis=1)) / df2[cc].std(ddof=0, axis=1)) < std
        
        # # # # replace outliers with nan
        # df2[cc]= df2[cc].where(lim, np.nan)
        
        
        
        # df2[cc] = df2[cc].astype(float).apply(lambda x: x.interpolate(method='linear', limit_direction='both'), axis=1).ffill().bfill()
       
        
        # df2=df2.T
        
        # #find outliers using z-score iter 2
        # lim = np.abs((df2 - df2.mean(axis=0)) / df2.std(ddof=0,axis=0)) < std
        # #lim=df2.apply(stats.zscore, axis=1) <1
        # # # # replace outliers with nan
        # df2= df2.where(lim, np.nan)
        
        # df2= df2.astype(float).apply(lambda x: x.interpolate(method='linear', limit_direction='both'), axis=0).ffill().bfill()
        
        # for col in df2.columns:
        #     #print (col)
        #     #df2[col]=pd.to_numeric(df2[col])
        #     df2[col]= df2[col].interpolate(method='index', axis=0).ffill().bfill()
        
        # df2=df2.T
            
           
        #Add Timesereises column names
        df2.columns = dnames
        
        df2 = dd.from_pandas(df2, npartitions=10)
        
        
        # define a function to replace outliers with NaN using z-scores along each row
        def replace_outliers(row, stdd=std):
            zscores = (row - row.mean()) / row.std()
            row[abs(zscores) > stdd] = np.nan
            return row

        # apply the function to each row using apply
        df2 = df2.apply(replace_outliers, axis=1)
        
        #df2=df2.compute()
        
        # Select columns with 'float64' dtype  
        #float64_cols = list(df2.select_dtypes(include='float64'))

        # The same code again calling the columns
        df2[dnames] = df2[dnames].astype('float32')
        
        
        
        df2[dnames] = df2[dnames].apply(lambda x: x.interpolate(method='linear', limit_direction='both'), axis=1).ffill().bfill()
        
        df2=df2.compute()
        
        df2=df2.T
        for col in df2.columns:
            #print (col)
            #df2[col]=pd.to_numeric(df2[col])
            df2[col]= df2[col].interpolate(method='index', axis=0).ffill().bfill()
        
        df2=df2.T
        
        df2.columns = dnames
        
        
        

        # # interpolate missing values along each row
        # df2.interpolate(axis=1, limit_direction='both', limit_area='inside', method='linear', inplace=True)
                
        #  # Forward fill the DataFrame
        # df2.ffill(inplace=True)

        # # Backward fill the DataFrame
        # df2.bfill(inplace=True)
        
        #Calculate Linear Velocity for each data point
        def linear_VEL(df, dnames):
            
            # def best_fit_slope_and_intercept(xs,ys):
            #     from statistics import mean
            #     xs = np.array(xs, dtype=np.float64)
            #     ys = np.array(ys, dtype=np.float64)
            #     m = (((mean(xs)*mean(ys)) - mean(xs*ys)) /
            #         ((mean(xs)*mean(xs)) - mean(xs*xs)))
                
            #     b = mean(ys) - m*mean(xs)
                
            #     return m, b
            dd_list=[x.replace("D", "") for x in dnames]
            dates_list=([datetime.strptime(x, '%Y%m%d') for x in dd_list])
            days_num=[( ((x) - (pd.Timestamp(year=x.year, month=1, day=1))).days + 1) for x in dates_list]
            days_num=list(range(0, len(dnames)))
            dslope=[]
            std_slope=[]
            for index, dr in df.iterrows():
                #if index==0:
                rows=df.loc[index, :].values.flatten().tolist()
                row_values=rows
                # dfr = pd.DataFrame(dr).transpose()
                # dfr = dfr.loc[:, ~dfr.columns.str.contains('^Unnamed')]
            
                #slopeVEL=best_fit_slope_and_intercept(days_num, row_values)
                #print("slope", slopeVEL[0])
                slope, intercept, r_value, p_value, std_err = stats.linregress(days_num, row_values)
                dslope.append(slope)
                std_slope.append(std_err)
            return dslope, std_slope
        
        
        
        
            
        
        ###########################################################################
  
        
        dnames_new=[x.replace("D", "") for x in dnames]
        def input_dates(start_date="YYYYMMDD", end_date="YYYYMMDD"):
            start_date1=parser.parse(start_date)
            end_date2=parser.parse(end_date)
            date_list_start=[]
            date_list_end=[]
            for idx, item in enumerate(dnames_new):
                #filepath1, img_name = os.path.split(item) 
                str_date1=item
                str_date2=dnames_new[len(dnames_new)-1]
                #input start date
                date_time1 = parser.parse(str_date1)
                date_list_start.append(date_time1)
                #input end date
                date_time2 = parser.parse(str_date2)
                date_list_end.append(date_time2)
            st_date=min(date_list_start, key=lambda d: abs(d - start_date1))
            text_date1=st_date.strftime("%Y%m%d")
            End_date=min(date_list_end, key=lambda d: abs(d - end_date2))
            No_ofDays=(End_date-st_date).days
            
            text_date2=End_date.strftime("%Y%m%d")
            return [text_date1, text_date2, No_ofDays]

        velocity_scale=(input_dates(start_date=dnames_new[0], end_date=dnames_new[len(dnames_new)-1]))
        
        #################################
        # for idx, row in df2[dnames].iterrows():
        #     lim = np.abs((row[dnames] - df2[dnames]()) / row[dnames].std(ddof=0)) < 1
        #     row[dnames]= row[dnames].where(lim, np.nan)
        #     row[dnames] = row[dnames].astype(float).apply(lambda x: x.interpolate(method='linear', limit_direction='both'), axis=1).ffill().bfill()
            
        
        print (df2.describe())
        temp_df=pd.DataFrame()
        temp_df[dnames[0]]=df2[dnames[0]]
        #Choosing first date as reference for Time Series
        
        if master_reference==True:
            
            df2 = df2.sub(df2[dnames[0]], axis=0)
        else:
            
            df2=df2.diff(axis = 1, periods = 1)
        # count=0
        # for idx, col in enumerate(df2.columns):
        #     df2[col] = df2[col].sub(df2[dnames[count]], axis=0)
        #     count=count+1
            
       
        df2[dnames[0]]=0
            
        linear_velocity=linear_VEL(df2[dnames], dnames)
        out=df2
        if velocity_mode=="mean":
            out['VEL']=out[dnames].mean(axis=1)
            out['VEL_STD']=out[dnames].std(axis=1)
        elif velocity_mode=="linear":
            out['VEL']=linear_velocity[0]
            out['VEL_STD']=linear_velocity[1]
        if VEL_Scale=="month": 
            out['VEL']=out['VEL']/velocity_scale[2] * 30  #velocity_scale[2] is number of days
            out['VEL_STD']=out['VEL_STD'] /velocity_scale[2] *30
        elif VEL_Scale=="year":
            out['VEL']=out['VEL']/velocity_scale[2] * 365
            out['VEL_STD']=out['VEL_STD']/velocity_scale[2] * 365
        else:
            out['VEL']=out['VEL']
            out['VEL_STD']=out['VEL_STD']
               
        
            
        out['geometry']=df['geometry']
        out['CODE']=df['SiteID']
        #out[dnames[0]]=temp_df[dnames[0]]
        # out['HEIGHT']=0
        # out['H_STDEV']=0
        #out['V_STDEV']=out[dnames].std(axis=1)
        #out['COHERENCE']=0
        #out['H_STDEF']=0
        out['x']=df['x']
        out['y']=df['y']

        col_titles=['CODE','geometry','x', 'y', 'VEL', 'VEL_STD' ]+dnames
        out = out.reindex(columns=col_titles)
        
        

        geo_out=gpd.GeoDataFrame(out, geometry='geometry', crs=df.crs)

        geo_out.to_file(output_folder +"/" + outputFilename)
        (geo_out)

        return geo_out, dnames, linear_VEL
    
    if output_folder=="":
            output_folder= "stack_data/TS"
            
    
    if not os.path.exists(output_folder):
                os.makedirs(output_folder)
    if outputFilename=="":
            outputFilename= "TS_2D_"+ os.path.basename(velocity_points)
            
            
            
    EW=Helper_Time_Series(stacked_raster=stacked_raster_EW, velocity_points=velocity_points ,
                             dates_name=dates_name, output_folder=output_folder, outputFilename="TS_EW_"+ os.path.basename(velocity_points), std=std, VEL_Scale=VEL_Scale)
                             
    NS=Helper_Time_Series(stacked_raster=stacked_raster_NS, velocity_points=velocity_points, 
                             dates_name=dates_name, output_folder=output_folder, outputFilename="TS_NS_"+ os.path.basename(velocity_points), std=std, VEL_Scale=VEL_Scale)
    
    if not os.path.exists(output_folder):
                os.makedirs(output_folder)
    if outputFilename=="":
            outputFilename= "TS_2D_"+ os.path.basename(velocity_points)
            
            
    gdf_ew=EW[0]
    gdf_ns=NS[0]
    dnames=NS[1]
    df_2D_VEL=pd.DataFrame()
    df_2D_VEL['CODE']=gdf_ew['CODE']
    df_2D_VEL['geometry']=gdf_ew['geometry']
    df_2D_VEL['x']=gdf_ew['x']
    df_2D_VEL['y']=gdf_ew['y']
    
   
   #Calculate resultant velocity magnitude
    for col in gdf_ew[dnames].columns:
       
        df_2D_VEL[col]=np.hypot(gdf_ns[col],gdf_ew[col])
       
    df_2D_VEL['VEL_MEAN']=df_2D_VEL[dnames].mean(axis=1)
    df_2D_VEL['V_STDEV']=df_2D_VEL[dnames].std(axis=1)
    #we call linear velocity function from above then reuse it below to replace VEL_2D Mean an STD below for lines
    # linear_2D_Velocity_function=EW[2]
    # linear_2D_Velocity=linear_2D_Velocity_function(df_2D_VEL[dnames], dnames)
    # df_2D_VEL['VEL']=linear_2D_Velocity[0]
    # df_2D_VEL['V_STDEV']=linear_2D_Velocity[1]
    #############################
    col_titles=['CODE','geometry','x', 'y', 'VEL_MEAN' , 'V_STDEV' ]+ dnames 
    df_2D_VEL = df_2D_VEL.reindex(columns=col_titles)
    gdf_2D_VEL=gpd.GeoDataFrame(df_2D_VEL, geometry='geometry', crs=gdf_ew.crs)
    
    
    
    gdf_2D_VEL.to_file(output_folder +"/" + outputFilename)
    
    
    #Calculate resultant velocity direction
    
    dir_df_2D_VEL=pd.DataFrame()
    dir_df_2D_VEL['CODE']=gdf_ew['CODE']
    dir_df_2D_VEL['geometry']=gdf_ew['geometry']
    dir_df_2D_VEL['x']=gdf_ew['x']
    dir_df_2D_VEL['y']=gdf_ew['y']
    
    newcol_dir_list=[]
    for col in gdf_ew[dnames].columns:
        newcol_dir= col
        newcol_dir_list.append(newcol_dir)
        dir_df_2D_VEL[newcol_dir]=np.arctan2(gdf_ns[col],gdf_ew[col])
        dir_df_2D_VEL[newcol_dir]=np.degrees(dir_df_2D_VEL[newcol_dir])
        dir_df_2D_VEL[newcol_dir]=(450-dir_df_2D_VEL[newcol_dir]) % 360
    dir_df_2D_VEL['VELDir_MEAN']=dir_df_2D_VEL[newcol_dir_list].mean(axis=1)
    col_titles=['CODE','geometry','x', 'y', 'VELDir_MEAN'  ]+ newcol_dir_list
    dir_df_2D_VEL = dir_df_2D_VEL.reindex(columns=col_titles)
    dir_gdf_2D_VEL=gpd.GeoDataFrame(dir_df_2D_VEL, geometry='geometry', crs=gdf_ew.crs)
    
    dir_gdf_2D_VEL.to_file(output_folder +"/" + outputFilename[:-4]+"_dir.shp")
    
    #Calcuate Mean Corrected velocity products MEAN X, Y, 2D and Dir
    corrected_mean_products=pd.DataFrame()
    corrected_mean_products['CODE']=gdf_ew['CODE']
    corrected_mean_products['geometry']=gdf_ew['geometry']
    corrected_mean_products['x']=gdf_ew['x']
    corrected_mean_products['y']=gdf_ew['y']
    corrected_mean_products['VEL_E']=gdf_ew['VEL']
    corrected_mean_products['VEL_N']=gdf_ns['VEL']
    #corrected_mean_products['VEL_2D']=df_2D_VEL['VEL_MEAN']
    corrected_mean_products['VEL_2D']=df_2D_VEL['VEL_MEAN']
    corrected_mean_products['2DV_STDEV']=df_2D_VEL['V_STDEV']
    corrected_mean_products['VEL_2DDir']=dir_df_2D_VEL['VELDir_MEAN']
    corrected_mean_products_geo=gpd.GeoDataFrame(corrected_mean_products, geometry='geometry', crs=gdf_ew.crs)
    
    corrected_mean_products_geo.to_file(output_folder +"/" + outputFilename[:-4]+"_mean.shp")
    
    

def akhdefo_dashApp(directory="", DEM_Path='', std_thresh_interpolation=1.0, port=8051, smoothing=False, hillshade_option=True, normalize_cmap=False):


    """
    An application that visualizes and analyzes raster data using Dash.

    Args:
        directory (str): The path to the directory containing the raster files. Default is an empty string.
        DEM_Path (str): The path to the DEM (Digital Elevation Model) file. Default is an empty string.
        std_thresh_interpolation (float): The standard deviation threshold for interpolation. Default is 1.0.
        port (int): The port number for running the Dash server. Default is 8051.
        smoothing(bool): if True re-interpolate the entire raster with 16 pixel average to fill Nan Values 

    Returns:
        Dash: The Dash application object.

    """
    import os
    from datetime import datetime

    import cmocean
    import dash
    import dash_daq as daq
    import earthpy.spatial as es
    import matplotlib.cm as cm
    import matplotlib.pyplot as plt
    import numpy as np
    import plotly.colors as colors
    import rasterio
    from dash import dcc, html
    from dash.dependencies import Input, Output, State
    from rasterio.enums import Resampling
    from rasterio.transform import from_origin
    from scipy import interpolate
    
    
    import os
    import glob
    from osgeo import gdal

    def get_smallest_dimensions(folder_path):
        """
        Get the smallest width and height among all rasters in the folder.
        """
        smallest_width = float('inf')
        smallest_height = float('inf')

        # Loop through each raster file in the folder
        for raster_file in glob.glob(os.path.join(folder_path, '*.tif')):
            ds = gdal.Open(raster_file, gdal.GA_ReadOnly)
            
            if ds is None:
                print(f"Unable to open {raster_file}")
                continue

            width = ds.RasterXSize
            height = ds.RasterYSize

            #print(f"{raster_file} Dimensions: {width} x {height}")
            
            if width < smallest_width:
                smallest_width = width
            if height < smallest_height:
                smallest_height = height

            ds = None  # Close the dataset

        return smallest_width, smallest_height

    def crop_raster(input_file, target_width, target_height):
        """
        Crop a raster to the specified dimensions without resampling and overwrite the original.
        """
        # Temporary output file
        temp_output = input_file + '_temp.tif'
        
        options = gdal.TranslateOptions(format='GTiff', width=target_width, height=target_height)
        ds = gdal.Open(input_file)
        if ds:
            gdal.Translate(temp_output, ds, options=options)
            ds = None  # Close the dataset
            
            # Overwrite the original raster with the cropped version
            os.remove(input_file)
            os.rename(temp_output, input_file)
        else:
            print(f"Failed to open {input_file}")


    

    # Step 1: Get the smallest raster dimensions
    smallest_width, smallest_height = get_smallest_dimensions(directory)
    #print(f"Target Dimensions: {smallest_width} x {smallest_height}")

    # Step 2: Crop each raster to the smallest dimensions
    for raster_file in glob.glob(os.path.join(directory, '*.tif')):
        crop_raster(raster_file, smallest_width, smallest_height)


    app = dash.Dash(__name__)
    
    
    
    def stack_rasters(directory):
        from scipy.ndimage import zoom
        src_dem=rasterio.open(DEM_Path)
        template=src_dem.read(1)
        
        raster_array_list = []
        transform_list=[]
        raster_files = sorted([f for f in os.listdir(directory) if f.endswith('.tif')])
        
        for file in raster_files:
            with rasterio.open(os.path.join(directory, file)) as src:
                raster_array=src.read(1)
                raster_array_list.append(raster_array)
                transform_list.append(src.transform)
                crs=src.crs
        stacked_array_init=np.dstack(raster_array_list)
        stacked_array=np.dstack(raster_array_list)
        # # Set the first layer to zero
        # stacked_array[:, :, 0] = 0       
        # # Update each band in the raster stack array
        # for idx in range(1, stacked_array.shape[2]): # This iterates over the layers
        #     stacked_array[:, :, idx] += stacked_array[:, :, idx - 1]
            
        return stacked_array , raster_files, transform_list, crs, stacked_array_init

    def calculate_mean_circle(raster_stack, x, y, r):
        indices = np.indices(raster_stack.shape[:2])
        distances = np.sqrt((indices[0] - y)**2 + (indices[1] - x)**2)
        mask = distances < r
        pixel_values = raster_stack[mask, :]
        return np.nanmean(pixel_values, axis=0)
    
    def save_layers_as_geotiff(raster_stack, transforms, crs, output_dir, output_prefix, raster_files):

        ################Uncomment this section if you want to subtrcat each raster deformation to first raster deformation)###################

        # # Create a copy of the raster stack
        # result_stack = np.copy(raster_stack)
        # # Create an empty mask array with the same shape as the raster stack
        # mask_stack = np.zeros_like(raster_stack)
        
        # # Subtract each band by the first band
        # for i in range((result_stack.shape[2])):
        #     mask = np.isnan(raster_stack[:, :, i])  # Create a mask for NaN values
        #     result_stack[:, :, i] = np.where(mask, 0, raster_stack[:, :, i])  # Replace NaN values with zero
        #     result_stack[:, :, i]=result_stack[:, :, i]-result_stack[:, :, 0]
        #     mask_stack[:, :, i] = mask  # Store the mask for this band in the mask stack
        #     result_stack[:, :, i][mask]=np.nan
        
        # raster_stack=result_stack

        #########################################################################################################################################

        path=output_dir

        import re

        def convert_path(path):
            if re.match(r'^[A-Za-z]:\\', path):
                path = r'\\'.join(re.split(r'[\\/]+', path))
            return path

        
        converted_path = convert_path(path)
        converted_path=converted_path + "/" + "corrected"

        if not os.path.exists(converted_path):
            os.makedirs(converted_path)

       

        base_names = [os.path.basename(file) for file in raster_files]
        count=0
        for layer_index in range(raster_stack.shape[2]):
            output_filename = output_prefix + "_" + base_names[count]
            output_path = os.path.join(converted_path, output_filename)

            # Create a new GeoTIFF file
            with rasterio.open(output_path, 'w', driver='GTiff', height=raster_stack.shape[0], 
                            width=raster_stack.shape[1], count=1, dtype=raster_stack.dtype,
                            crs=crs, transform=transforms[layer_index]) as dst:
                # Write the layer to the GeoTIFF file
                dst.write(raster_stack[:, :, layer_index], 1)

            print(f"Layer {layer_index} saved as {output_path}")
            count=count+1
        

    def interpolate_rasters(raster_stack, raster_files, transform_list, crs, std_thresh_interpolation, smoothing):

        #raster_stack
        
        # #raster_stack=raster_stack - raster_stack[0]
        # # Calculate the cumulative sums along the third dimension
        # #raster_stack = np.cumsum(raster_stack, axis=2)



        if std_thresh_interpolation is not None:

            x = np.arange(0, raster_stack.shape[2])
            y = np.arange(0, raster_stack.shape[1])
            z = np.arange(0, raster_stack.shape[0])

             
            #  # Replace outliers with NaN
            outlier_mask = np.abs(raster_stack - np.nanmean(raster_stack)) > std_thresh_interpolation * np.nanstd(raster_stack)
            raster_stack[outlier_mask] = np.nan

            interp_func = interpolate.RegularGridInterpolator((z, y, x), raster_stack, method='nearest')

            nan_locs = np.where(np.isnan(raster_stack))
            raster_stack[nan_locs] = interp_func(nan_locs)
        
        #  # Replace outliers with NaN
        # outlier_mask = np.abs(raster_stack - np.nanmean(raster_stack)) > std_thresh_interpolation * np.nanstd(raster_stack)
        # raster_stack[outlier_mask] = np.nan

        # #################
        # #Reinterpolate non-NaN values with the 16-pixel average of nearest values. This helps to elimnate spikes in the dataset
        if smoothing==True:
           
            from scipy.ndimage import convolve

            def replace_non_nan_with_avg(arr):
                kernel = np.ones((3, 3, 1))
                kernel[1, 1, 0] = 0  # Exclude the center pixel from the kernel

                # Calculate the sum of the neighboring pixels for each non-NaN pixel
                sum_neighbors = convolve(arr, kernel, mode='constant', cval=np.nan)

                # Count the number of non-NaN neighbors for each pixel
                non_nan_count = convolve(~np.isnan(arr).astype(int), kernel, mode='constant', cval=0)

                # Calculate the average of neighboring pixels and replace non-NaN values with the average
                avg_neighbors = np.where(non_nan_count > 0, sum_neighbors / non_nan_count, np.nan)
                filled_arr = np.where(np.isnan(arr), avg_neighbors, arr)

                return filled_arr
            raster_stack = replace_non_nan_with_avg(raster_stack)   
            # raster_stack = np.where(non_nan_count > 0, np.divide(raster_sum, non_nan_count, where=non_nan_count > 0), np.nan)


        return raster_stack, raster_files, transform_list, crs
    
  
    from datetime import datetime

    def calculate_total_days(start_date, end_date):
        start_date_obj = datetime.strptime(start_date, "%Y-%m-%d")
        end_date_obj = datetime.strptime(end_date, "%Y-%m-%d")

        time_difference = end_date_obj - start_date_obj
        total_days = time_difference.days + 1

        return total_days

    def calculate_cumulative_deformation(lst):
        lst = np.array(lst)
        cumulative_deformation = np.insert(np.cumsum(lst), 0, 0)
        #cumulative_deformation=np.cumsum(lst)
        return cumulative_deformation.tolist()

    def calculate_std(pixel_values):
        std = np.std(pixel_values)
        return std
    from numpy import arange, where, zeros, zeros_like
    from scipy.interpolate import interp1d
    from scipy.stats import linregress

    def interpolate_nans(band_values):
        mask = np.isnan(band_values)
        if mask.all():
            return np.zeros_like(band_values)
        else:
            x = np.arange(len(band_values))
            y = np.array(band_values)[~mask]
            f = interp1d(x[~mask], y, bounds_error=False, fill_value='extrapolate')
            xnew = np.arange(len(band_values))
            return f(xnew)
    
    def calculate_linear_change(pixel_values, polyfit_order):
       
        times = arange(len(pixel_values))
        # Fit a linear polynomial (degree 1) to the data
        coefficients = np.polyfit(times, pixel_values, polyfit_order)
        # Extract the slope and intercept from the coefficients
        slope = coefficients[0]
        # Calculate the residuals (errors)
        residuals = pixel_values - (slope * times + coefficients[1])
        # Calculate the standard deviation of the residuals
        std_res = np.nanstd(residuals)
        std=np.nanstd(pixel_values)

        return slope, std, residuals
    
    # def create_linear_velocity_map(raster_stack):
    #     height, width, num_bands = raster_stack.shape

    #     velocity_map = np.zeros((height, width), dtype=float)

    #     band_values = np.transpose(raster_stack, axes=(2, 0, 1)).reshape(num_bands, -1)

    #     #interpolated_band_values = np.apply_along_axis(interpolate_nans, axis=0, arr=band_values)

    #     times = np.arange(num_bands)

    #     slopes = np.apply_along_axis(lambda x: linregress(times, x)[0], axis=0, arr=band_values)

    #     velocity_map = slopes.reshape(height, width)

    #     return velocity_map
    
    def create_cumulative_deformation_map(raster_stack):
        # # Extract the dimensions of the raster stack
        # height, width, num_bands = raster_stack.shape

        # # Create an empty array to store the cumulative deformation map
        # deformation_map = np.zeros((height, width), dtype=float)

        #     # Get the values of each band for all pixels at once
        # band_values = raster_stack.reshape(-1, raster_stack.shape[2])

        # Compute the mean deformation 
        deformation_map = np.nanmean(raster_stack, axis=2)

        # Reshape the deformation array to match the dimensions of deformation_map
        #deformation_map = deformation.reshape(height, width)
        

        return deformation_map

    
    def interpolate_Listnan_values(data_list):
        # Convert list to numpy array
        arr = np.array(data_list)
        
        # Indices of the array
        ind = np.arange(arr.shape[0])
        
        # Boolean array indicating the presence of NaN values
        mask = np.isnan(arr)
        
        # Interpolate NaN values
        arr[mask] = np.interp(ind[mask], ind[~mask], arr[~mask])
        
        # Propagate the first non-NaN value to NaN values at the start (if any)
        for i in range(len(arr)):
            if not np.isnan(arr[i]):
                arr[:i] = arr[i]
                break

        # Propagate the last non-NaN value to NaN values at the end (if any)
        for i in range(len(arr) - 1, -1, -1):
            if not np.isnan(arr[i]):
                arr[i + 1:] = arr[i]
                break
        
        return arr.tolist()
    
    
    def calculate_mean_raster(raster_stack):
        # Ensure rasters is a numpy array
        raster_stack = np.array(raster_stack)

        mean_raster = np.nanmean(raster_stack, axis=0)
        
        
        
        return mean_raster
    
    def export_mean_raster(mean_raster, crs, transform, output_directory):
    # Create a new GeoTIFF file for the mean raster
        path=output_directory

        import re

        def convert_path(path):
            if re.match(r'^[A-Za-z]:\\', path):
                path = r'\\'.join(re.split(r'[\\/]+', path))
            return path

        
        converted_path = convert_path(path)

        if not os.path.exists(converted_path):
            os.makedirs(converted_path)

        

        output_path = os.path.join(converted_path, "Mean_Annual_VEL.tif")
        with rasterio.open(output_path, 'w', driver='GTiff', height=mean_raster.shape[0],
                        width=mean_raster.shape[1], count=1, dtype=mean_raster.dtype,
                        crs=crs, transform=transform) as dst:
            dst.write(mean_raster, 1)
        print("Mean raster saved as GeoTIFF: ", output_path)
        return 
        
    raster_directory = directory
    raster_stack, raster_files, transform_list, crs, stacked_array_init= stack_rasters(raster_directory)
    
    raster_stack, raster_files, transform_list, crs = interpolate_rasters( raster_stack, raster_files, transform_list, crs, std_thresh_interpolation=std_thresh_interpolation, smoothing=smoothing)
    
    Mean_Cumulative_Deformation = create_cumulative_deformation_map(raster_stack)
    Mean_deformation_map = create_cumulative_deformation_map(stacked_array_init)
    
    
    
    meandefo_vmin=np.nanmin(Mean_deformation_map)
    meandefo_vmax=np.nanmax(Mean_deformation_map)
    
    zmin_data = meandefo_vmin
    zmax_data = meandefo_vmax
    zrange = zmax_data - zmin_data
    step_size = zrange / 10
    
    def generate_slider_marks_positive(zmax, n_marks=10):
        """Generate 10 marks for positive values up to zmax."""
        step = zmax / (n_marks - 1)
        return {str(round(i * step, 2)): str(round(i * step, 2)) for i in range(n_marks)}

    def generate_slider_marks_negative(zmin, n_marks=10):
        """Generate 10 marks for negative values down to zmin."""
        step = abs(zmin) / (n_marks - 1)
        return {str(round(-i * step, 2)): str(round(-i * step, 2)) for i in range(n_marks)}
    if normalize_cmap==True:
        scale_value_mean_defo= (meandefo_vmin,meandefo_vmax ) 
        maximum_value_defo = abs(scale_value_mean_defo[0]) if abs(scale_value_mean_defo[0]) > abs(scale_value_mean_defo[1]) else abs(scale_value_mean_defo[1])
        meandefo_vmin=-maximum_value_defo
        meandefo_vmax=maximum_value_defo
    
    #linear_velocity_map = create_linear_velocity_map(raster_stack)
    
    #mean_raster, raster_diff = calculate_mean_raster(raster_stack)

    raster_dates = [datetime.strptime(f[:8], "%Y%m%d") for f in raster_files]
    dem_path = DEM_Path
    with rasterio.open(dem_path) as dem_src:
        dem = dem_src.read(1)

    if hillshade_option==True:

        hillshade = es.hillshade(dem)
    else: 
        hillshade=dem
    # import matplotlib.cm as cm

    # # Get a list of all available colormaps
    # cmap_options = [cmap for cmap in cm.datad]

    # # Set the default cmap value
    # default_cmap = cmap_options[0]
    # # Get cmocean colormap names
    # cmaps = cmocean.cm.cmapnames


    def matplotlib_to_plotly(cmap, pl_entries):
        h = 1.0/(pl_entries-1)
        pl_colorscale = []
        
        for k in range(pl_entries):
            C = list(map(np.uint8, np.array(cmap(k*h)[:3])*255))
            pl_colorscale.append([k*h, 'rgb'+str((C[0], C[1], C[2]))])
            
        return pl_colorscale

        # Include all cmocean colormaps
    cmap_options = [cmap for cmap in cmocean.cm.cmapnames]
    # Include all matplotlib colormaps
    cmap_options.extend([cmap for cmap in plt.colormaps()])

        # Define the app layout
    app.layout = html.Div([
        html.Div([
            html.Div([
                html.Label('X-axis Label:'),
                dcc.Input(id='x-axis-input', type='text', placeholder='Enter X-axis label', value='Dates'),
                html.Label('Y-axis Label:'),
                dcc.Input(id='y-axis-input', type='text', placeholder='Enter Y-axis label', value='cm'),
                html.Label('Enter Path to Save Geotif Products:'),
                dcc.Input(id='Path_meanraster-input', type='text', placeholder='Enter Path to Directory to Save Mean Velocity Raster', value='./'),
                html.Div(id="output-div"), html.Br(), html.Label("zmin"),
            dcc.Slider(
                id='zmin-slider',
                min=zmin_data,
                max=zmax_data,
                value=zmin_data,
                marks=generate_slider_marks_negative(zmin_data),
                step=step_size
            ),
            html.Br(),
            html.Label("zmax"),
            dcc.Slider(
                id='zmax-slider',
                min=zmin_data,
                max=zmax_data,
                value=zmax_data,
                marks=generate_slider_marks_positive(zmax_data),
                step=step_size
            ),
            ], style={'padding': '10px', 'vertical-align': 'top'}),
            
            html.Div([
                dcc.Graph(id='raster-plot', figure={
                    'data': [
                    {'z': hillshade, 'type': 'heatmap', 'colorscale': 'Greys', 'name': 'Hillshade',
                    'colorbar': {'thickness': 0}, 'layer': 'below'},  # Hide the colorbar for hillshade trace
                    {'z': Mean_deformation_map, 'type': 'heatmap', 'colorscale': 'Viridis', 'name': 'Deformation', 'zmin': meandefo_vmin, "zmax":meandefo_vmax,
                    'colorbar': {'title': 'colorbar-label'}}
                    ],
                    'layout': {
                        'title': 'Raster Overlay',
                        'hovermode': 'closest',
                        'yaxis': {'autorange': 'reversed'},
                        'margin': {'l': 50, 'r': 50, 't': 50, 'b': 50},
                        'height': '500px',
                        'resizable': True,
                        'colorway': ['black', 'blue'],
                        'coloraxis': {'colorbar': {'title': 'Colorbar Label'}}  # Add colorbar label
                    }
                })
            ], style={'width': '100%', 'display': 'inline-block', 'vertical-align': 'top'}),
        
            html.Div([
                dcc.Graph(id='profile-plot', style={'height': '400px'})
            ], style={'width': '100%', 'display': 'inline-block', 'vertical-align': 'top'})
        ], style={'padding': '10px', 'width': '60%', 'display': 'inline-block', 'vertical-align': 'top'}),
        
        html.Div([
            html.Div([
                html.Div([
                    dcc.RadioItems(
                        id='selection-type2',
                        options=[{'label': i, 'value': i} for i in ['Mean-Deformation', 'Mean-Cumulative-Deformation', 'Annual-Velocity']],
                        value='Mean-Deformation',
                        labelStyle={'display': 'inline-block'}
                    ),
                    html.Hr(),  # Add horizontal line
                    dcc.Input(id='fig_title', type='text', placeholder='Type Figure Title', value='Figure Title'),
                    html.Hr(),  # Add horizontal line
                    dcc.Input(id='fig_colorbar_label', type='text', placeholder='Type Figure Colorbar Label', value='Colorbar Label'),
                    html.Hr(),  # Add horizontal line
                    html.Label('Circle Size:'),
                    daq.Slider(
                        id='circle-size-slider',
                        min=1,
                        max=100,
                        step=1,
                        value=10,
                        marks={i: str(i) for i in range(0, 101, 10)},
                        included=True,
                    ),
                    html.Div(style={'height': '20px'})  # Add extra space
                ]),
                
                html.Div([
                    dcc.RadioItems(
                        id='selection-type',
                        options=[{'label': i, 'value': i} for i in ['Point', 'Circle']],
                        value='Point',
                        labelStyle={'display': 'inline-block'}
                    ),
                    html.Hr(),  # Add horizontal line
                    html.Label('Start Date:'),
                    html.Hr(),  # Add horizontal line
                    dcc.DatePickerSingle(
                        id='start-date-picker',
                        min_date_allowed=min(raster_dates),
                        max_date_allowed=max(raster_dates),
                        initial_visible_month=min(raster_dates),
                        date=min(raster_dates)
                    ),
                    html.Div(id="selected-start-date"),
                    html.Hr(),  # Add horizontal line
                    html.Label('End Date:'),
                    dcc.DatePickerSingle(
                        id='end-date-picker',
                        min_date_allowed=min(raster_dates),
                        max_date_allowed=max(raster_dates),
                        initial_visible_month=max(raster_dates),
                        date=max(raster_dates)
                    ),
                    html.Div(id="selected-end-date"),
                    html.Hr(),  # Add horizontal line
                    html.Button('Reset', id='reset-button', n_clicks=0) ,
                ], style={'padding': '10px', 'vertical-align': 'top'})
            ], style={'width': '100%', 'display': 'inline-block', 'vertical-align': 'top'}),
            
            html.Div([
                html.Div([
                    html.Button('Toggle Trendline', id='toggle-trendline', n_clicks=0),
                    html.Hr(),
                    html.Button('Toggle Cumulative-Deformation', id='toggle-cum_sum', n_clicks=0),
                    html.Hr(),  # Add horizontal line
                    html.Button("Download GeoTIFF", id="btn_download", n_clicks=0),
                    dcc.Download(id="download"),
                    html.Label('Colormap:'),
                    html.Hr(),  # Add horizontal line
                    dcc.Dropdown(
                        id='colorscale-dropdown',
                        options=[{'label': cmap, 'value': cmap} for cmap in cmap_options],
                        value='Viridis',
                        clearable=True,
                        searchable=True
                    ),
                    html.Label('Transparency:'),
                    html.Hr(),  # Add horizontal line
                    daq.Slider(
                        id='transparency-slider',
                        min=0,
                        max=1,
                        step=0.1,
                        value=0.6,
                        marks={i / 10: str(i / 10) for i in range(0, 11)},
                        included=True , 
                    ), html.Div(style={'height': '20px'}),  # Add extra space
                    
                    html.Label('PolyFit-Order:', style={"margin-bottom": "10px"}),
                    
                    daq.Slider(
                        id='polyfit-slider',
                        min=1,
                        max=10,
                        step=1,
                        value=1,
                        marks={i : str(i) for i in range(1, 11, 1)},
                        included=True
                    )
                ], style={'padding': '20px', 'vertical-align': 'top'})
            ], style={'width': '100%', 'display': 'inline-block', 'vertical-align': 'top'})
        ], style={'padding': '10px', 'width': '10%', 'display': 'inline-block', 'vertical-align': 'top'})  # Adjust the width to 40%
    ], style={'padding': '10px', 'width': '100%', 'display': 'flex'})


        
   
    @app.callback(
        Output('raster-plot', 'figure'),
        [Input('colorscale-dropdown', 'value'), Input('transparency-slider', 'value'), Input('raster-plot', 'clickData'),
        Input('reset-button', 'n_clicks'), Input('selection-type', 'value'), Input('circle-size-slider', 'value'), Input('fig_title', 'value'),
        Input('fig_colorbar_label', 'value'), Input('start-date-picker', 'date'),
        Input('end-date-picker', 'date'), Input('selection-type2', 'value'), Input('zmin-slider', 'value'),
        Input('zmax-slider', 'value')]
    )
    def update_raster_plot(cmap_value, transparency, click_data, reset_clicks, selection_type, circle_size, Fig_title, Fig_cbar_label, start_date, end_date, selection_type2,
                           zmin, zmax):
        
        # start_index = raster_dates.index(datetime.strptime(start_date, "%Y-%m-%d"))
        # end_index = raster_dates.index(datetime.strptime(end_date, "%Y-%m-%d"))

        # raster_stack = raster_stack[:, :, start_index:end_index + 1]

       
        if cmap_value in cmocean.cm.cmapnames:
            cmap = cm.get_cmap('cmo.'+cmap_value.lower())
        else:
            cmap = cm.get_cmap(cmap_value.lower())
            
        cmap_plotly = matplotlib_to_plotly(cmap, 255)
        
        
        
        if reset_clicks:
            return {
                'data': [ {'z': hillshade, 'type': 'heatmap', 'colorscale': 'Greys', 'zmin': 0, 'zmax': 255, 'name': 'Hillshade', 'colorbar': {'thickness': 0}, 'layer': 'below'},
            {'z': Mean_deformation_map, 'type': 'heatmap', 'colorscale': cmap_plotly, 'name': 'Mean Deformation', 'opacity': transparency,
              'zmin': np.nanmin(Mean_deformation_map),
              'zmax': np.nanmax(Mean_deformation_map)}],
                'layout': {
                    'title': Fig_title,
                    'hovermode': 'closest',
                    'yaxis': {'autorange': 'reversed'},
                    'margin': {'l': 50, 'r': 50, 't': 50, 'b': 50},
                    'height': '400px',
                    'resizable': True,
                    'colorway': ['black', 'blue'], 'colorbar': {'title': Fig_cbar_label}  # Add colorbar label
                }
            }

       
        
        
        if selection_type2=='Annual-Velocity':
            
            total_days = calculate_total_days(start_date, end_date)
            total_days=int(total_days)
            Aver_annual_VEL_Map=Mean_deformation_map/total_days * 365
            
            
            vel_vmin=np.nanmin(Aver_annual_VEL_Map)
            vel_vmax=np.nanmax(Aver_annual_VEL_Map)
            
            if normalize_cmap==True:
                scale_value_vel = (vel_vmin, vel_vmax)
                maximum_value = abs(scale_value_vel[0]) if abs(scale_value_vel[0]) > abs(scale_value_vel[1]) else abs(scale_value_vel[1])

                vel_vmin=-maximum_value
                vel_vmax=maximum_value
            
           
            fig_data = [ {'z': hillshade, 'type': 'heatmap', 'colorscale': 'Greys', 'zmin': 0, 'zmax': 255, 'name': 'Hillshade', 'colorbar': {'thickness': 0}, 'layer': 'below'},
            {'z': Aver_annual_VEL_Map, 'type': 'heatmap', 'colorscale': cmap_plotly, 'name': 'Mean-VEL', 'opacity': transparency, 'zmin': zmin/total_days * 365,
              'zmax': zmax/total_days * 365}]
            
            layout = {
                'title': Fig_title,
                'hovermode': 'closest',
                'yaxis': {'autorange': 'reversed'},
                'margin': {'l': 50, 'r': 50, 't': 50, 'b': 50},
                'height': '400px',
                'resizable': True,
                'colorway': ['black', 'blue'], 'colorbar': { 'title': Fig_cbar_label}  # Add colorbar label

            }
            
        elif selection_type2=='Mean-Cumulative-Deformation':
            
            cumeandefo_vmin=np.nanmin(Mean_Cumulative_Deformation)
            cumeandefo_vmax=np.nanmax(Mean_Cumulative_Deformation)
            
            if normalize_cmap==True:
                scale_value_cumean_defo= max(abs(Mean_Cumulative_Deformation.min()), Mean_Cumulative_Deformation.max())  # or another scale as per your needs
                cumeandefo_vmin=-scale_value_cumean_defo
                cumeandefo_vmax=scale_value_cumean_defo
            
            total_days = calculate_total_days(start_date, end_date)
            total_days=int(total_days)
            
            fig_data = [ {'z': hillshade, 'type': 'heatmap', 'colorscale': 'Greys', 'zmin': 0, 'zmax': 255, 'name': 'Hillshade', 'colorbar': {'thickness': 0}, 'layer': 'below'},
            {'z': Mean_Cumulative_Deformation, 'type': 'heatmap', 'colorscale': cmap_plotly, 'name': 'Cummul-Defo', 'opacity': transparency, 'zmin': cumeandefo_vmin,
              'zmax': cumeandefo_vmax}]
            
            layout = {
                'title': Fig_title,
                'hovermode': 'closest',
                'yaxis': {'autorange': 'reversed'},
                'margin': {'l': 50, 'r': 50, 't': 50, 'b': 50},
                'height': '400px',
                'resizable': True,
                'colorway': ['black', 'blue'], 'colorbar': { 'title': Fig_cbar_label}  # Add colorbar label

            }
            
        elif selection_type2=='Mean-Deformation':
            
            meandefo_vmin=np.nanmin(Mean_deformation_map)
            meandefo_vmax=np.nanmax(Mean_deformation_map)

            if normalize_cmap==True:
                scale_value_mean_defo= (meandefo_vmin,meandefo_vmax ) 
                maximum_value_defo = abs(scale_value_mean_defo[0]) if abs(scale_value_mean_defo[0]) > abs(scale_value_mean_defo[1]) else abs(scale_value_mean_defo[1])
                meandefo_vmin=-maximum_value_defo
                meandefo_vmax=maximum_value_defo
            
            
        
            fig_data = [
                {'z': hillshade, 'type': 'heatmap', 'colorscale': 'Greys', 'zmin': 0, 'zmax': 255, 'name': 'Hillshade', 'colorbar': {'thickness': 0}, 'layer': 'below' },
                {'z': Mean_deformation_map, 'type': 'heatmap', 'colorscale': cmap_plotly, 'name': 'Mean-Defo', 'opacity': transparency, 'zmin': zmin,
                'zmax': zmax}

            ]
            

            layout = {
                'title': Fig_title,
                'hovermode': 'closest',
                'yaxis': {'autorange': 'reversed'},
                'margin': {'l': 50, 'r': 50, 't': 50, 'b': 50},
                'height': '400px',
                'resizable': True,
                'colorway': ['black', 'blue'], 'colorbar': {'title': Fig_cbar_label}  # Add colorbar label

            }

        if click_data:
            clicked_x = click_data['points'][0]['x']
            clicked_y = click_data['points'][0]['y']

            if selection_type == 'Point':
                fig_data.append({'x': [clicked_x], 'y': [clicked_y], 'mode': 'markers', 'marker': {'symbol': 'cross', 'color': 'red'}, 'name': 'Clicked Point'})
            elif selection_type == 'Circle':
                theta = np.linspace(0, 2*np.pi, 100)
                x = clicked_x + circle_size * np.cos(theta)
                y = clicked_y + circle_size * np.sin(theta)
                fig_data.append({'x': x, 'y': y, 'mode': 'lines', 'line': {'color': 'red'}, 'name': 'Circle'})
                layout['annotations'] = [{
                    'x': clicked_x,
                    'y': clicked_y,
                    'text': f"Circle (r={circle_size})",
                    'xref': 'x',
                    'yref': 'y',
                    'showarrow': False
                }]

        return {
            'data': fig_data,
            'layout': layout
        }

    @app.callback(Output("output-div", "n_clicks"),
            [
        Input('btn_download', 'n_clicks'), Input("Path_meanraster-input", "value"), Input('selection-type2', 'value'), Input('start-date-picker', 'date'),
        Input('end-date-picker', 'date')])
    
    def export_mean_raster_callback(n_clicks, path_saveFile, selection_type2, start_date, end_date ):
        
        if n_clicks > 0:
           
        
            if selection_type2=='Mean-Deformation':
                # cumulative_deformation_map = create_cumulative_deformation_map(raster_stack[start_index:end_index + 1])
                export_mean_raster(Mean_deformation_map, crs, transform_list[0], path_saveFile)
                save_layers_as_geotiff(raster_stack, transform_list, crs, path_saveFile, 'interpolated_', raster_files)
            elif selection_type2=='Annual-Velocity':
                # linear_velocity_map = create_linear_velocity_map(raster_stack[start_index:end_index + 1])
                total_days = calculate_total_days(start_date, end_date)
                total_days=int(total_days)
                Aver_annual_VEL_Map=Mean_deformation_map/total_days *360 
                export_mean_raster(Aver_annual_VEL_Map, crs, transform_list[0], path_saveFile)
                save_layers_as_geotiff(raster_stack, transform_list, crs, path_saveFile, 'interpolated_', raster_files)
        
    @app.callback(
        Output('profile-plot', 'figure'),
        [Input('raster-plot', 'clickData'), Input('toggle-trendline', 'n_clicks'), Input('start-date-picker', 'date'),
        Input('end-date-picker', 'date'), Input('reset-button', 'n_clicks'), Input('selection-type', 'value'),
        Input('circle-size-slider', 'value'),
        Input('x-axis-input', 'value'), Input('y-axis-input', 'value'), Input('toggle-cum_sum', 'n_clicks'), Input('polyfit-slider', 'value')]  # Add inputs for the textbox values
    )


    def update_profile_plot(click_data, n_clicks, start_date, end_date, reset_clicks, selection_type, circle_size, x_label, y_label, cum_sum, polyfit_order):
        
        def subtract_cumulative(data):
            cumulative_changes = [0]  # Initialize the cumulative changes with 0

            # Calculate the cumulative changes
            for i in range(1, len(data)):
                change = data[i] - data[i-1]
                cumulative_change = cumulative_changes[i-1] + change
                cumulative_changes.append(cumulative_change)
            return cumulative_changes
        
        if reset_clicks:
            return {}

        if click_data is None:
            return {}

        current_x = int(click_data['points'][0]['x'])
        current_y = int(click_data['points'][0]['y'])

        start_index = raster_dates.index(datetime.strptime(start_date, "%Y-%m-%d"))
        end_index = raster_dates.index(datetime.strptime(end_date, "%Y-%m-%d"))
        
       

        if selection_type == 'Point':
            pixel_values = raster_stack[current_y, current_x, start_index:end_index + 1]
            
            pixel_values=interpolate_Listnan_values(pixel_values)

            pixel_values_cumsum=subtract_cumulative(pixel_values)
            #pixel_values = [pixel_values[i] - pixel_values[0] for i in range(len(pixel_values))]
           

            #pixel_values = [pixel_values[i] - pixel_values[0] for i in range(len(pixel_values))]
            pixel_values_cumsum=np.cumsum(pixel_values_cumsum)

            
            total_days = calculate_total_days(start_date, end_date)
            total_days=int(total_days)
            #calculate days difference
            #total_days = end_index - start_index + 1



        elif selection_type == 'Circle':
            pixel_values = calculate_mean_circle(raster_stack[:, :, start_index:end_index + 1], current_x, current_y, circle_size)
            
            pixel_values=interpolate_Listnan_values(pixel_values)
            
            pixel_values_cumsum=subtract_cumulative(pixel_values)
           
            #pixel_values = [pixel_values[i] - pixel_values[0] for i in range(len(pixel_values))]
            pixel_values_cumsum=np.cumsum(pixel_values_cumsum)

            

            
            total_days = calculate_total_days(start_date, end_date)
            total_days=int(total_days)
         # Calculate slope, std, and residuals
        linear_change, std, residuals=calculate_linear_change(pixel_values, polyfit_order)
       

        std_mean = calculate_std(pixel_values)
        annual_slope=linear_change/total_days *365
        annual_std=std/total_days *365
        Aver_Defo=np.nanmean(pixel_values)
        Aver_VEL=Aver_Defo/total_days
        Aver_annual_VEL=Aver_Defo/total_days *365

        

        show_trendline = n_clicks % 2 != 0
        show_cum_sumProfile= cum_sum % 2 != 0

        profile_figure = {
            'data': [
                {'x': raster_dates[start_index:end_index + 1], 'y': pixel_values, 'mode': 'lines+markers', 'name': 'Deformation'}
            ],
            'layout': {
                'title': f'Linear-VEL: {annual_slope:.4f} {y_label}/year,  Std: {annual_std:.4f}, Ave-VEL:{Aver_annual_VEL:.4f}{y_label}/year , Ave-std: {std_mean:.4f}',
                'xaxis': {'title': x_label, 'tickangle': 45},
                'yaxis': {'title': y_label},
                'annotations': [
                    {
                        'x': 0.5,
                        'y': 1.21,
                        'xref': 'paper',
                        'yref': 'paper',
                        'text': f'Total-Days: {total_days}, Linear-VEL: {linear_change:.4f}{y_label}/{total_days}days, Std: {std:.4f}, Aver_Defo:{Aver_Defo:.4f}{y_label}',
                        'showarrow': False,
                        'font': {'color': 'black', 'size': 14}
                    }
                ]
            }
        }

        if show_trendline:
            
            x = np.arange(len(pixel_values))
            coefficients = np.polyfit(x, pixel_values, polyfit_order)
            polynomial = np.poly1d(coefficients)
            profile_figure['data'].append({'x': raster_dates[start_index:end_index + 1], 'y': polynomial(x), 'mode': 'lines', 'name': 'Trendline-Deformation'})
            # profile_figure['data'].append({'x': raster_dates[start_index:end_index + 1], 'y': residuals, 'mode': 'lines+markers', 'name': 'Residuals-Deformation'})
            
            #Residual Error and Trendline
        if show_cum_sumProfile:
           
           
           
            linear_change, std, residuals=calculate_linear_change(pixel_values_cumsum, polyfit_order)
            std_mean = calculate_std(pixel_values_cumsum)
            annual_slope=linear_change/total_days *365
            annual_std=std/total_days *365
            Aver_Defo=np.nanmean(pixel_values_cumsum)
            Aver_VEL=Aver_Defo/total_days
            Aver_annual_VEL=Aver_Defo/total_days *365
            x = np.arange(len(pixel_values_cumsum))
            coefficients = np.polyfit(x, pixel_values_cumsum, polyfit_order)
            polynomial = np.poly1d(coefficients)
            

        ###################################
           
            profile_figure = {
            'data': [
                {'x': raster_dates[start_index:end_index + 1], 'y': pixel_values_cumsum, 'mode': 'lines+markers',  'name': 'Cummulative-Deformation'}
                
            ],
            'layout': {
                'title': f'Linear-VEL: {annual_slope:.4f} {y_label}/year,  Std: {annual_std:.4f}, Ave-VEL:{Aver_annual_VEL:.4f}{y_label}/year , Ave-std: {std_mean:.4f}',
                'xaxis': {'title': x_label, 'tickangle': 45},
                'yaxis': {'title': y_label},
                'annotations': [
                    {
                        'x': 0.5,
                        'y': 1.21,
                        'xref': 'paper',
                        'yref': 'paper',
                        'text': f'Total-Days: {total_days}, Linear-VEL: {linear_change:.4f}{y_label}/{total_days}days, Std: {std:.4f}, Aver_Defo:{Aver_Defo:.4f}{y_label}',
                        'showarrow': False,
                        'font': {'color': 'black', 'size': 14}
                    }
                ]
            }
        }

            if show_trendline:
                
                x = np.arange(len(pixel_values_cumsum))
                coefficients = np.polyfit(x, pixel_values_cumsum, polyfit_order)
                polynomial = np.poly1d(coefficients)
                profile_figure['data'].append({'x': raster_dates[start_index:end_index + 1], 'y': polynomial(x), 'mode': 'line', 'name': 'Trendline-Cummulative'})
                # profile_figure['data'].append({'x': raster_dates[start_index:end_index + 1], 'y': residuals, 'mode': 'lines+markers', 'name': 'Residuals-Deformation'})
       
        ######################

        return profile_figure
    
    
    app.run_server(port=port)

    return app.run_server(port=port)



    


            

    



    

